package lab07;

import static jakarta.ws.rs.core.Response.Status.OK;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.nio.ByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.SocketChannel;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.Scanner;
import java.util.concurrent.CompletableFuture;

import jakarta.ws.rs.core.Response.Status;
import lombok.RequiredArgsConstructor;
import lombok.extern.log4j.Log4j2;

@Log4j2
@RequiredArgsConstructor
class HttpHandler {

	private final ByteBuffer buffer = ByteBuffer.allocate(1024*1024);
	
	private final Path root;
	
	private final StringBuilder requests = new StringBuilder();
	
	void handle(SocketChannel channel) {
		String request;
		try {
			request = getRequest(channel);
			requests.append(request);
			String resource = getResource(request);
			CompletableFuture<Void> future = transferFile(resource, channel);
			//task: #3
			future.handle((r,exc) -> { close(channel); return null;});
		} catch (IOException e) {
			writeStatus(Status.INTERNAL_SERVER_ERROR, channel);
			writeString("", channel);
		} catch (UnsupportedOperationException exc) {
			writeStatus(Status.METHOD_NOT_ALLOWED, channel);
			writeString("Method %s is not allowed".formatted(exc.getMessage()), channel);
		}
	}
	
	public void handleObserverChannel(SocketChannel accept) {
		writeStatus(OK, accept);
		writeString(requests.toString(), accept);
	}

	private void close(SocketChannel channel) {
		try {
			channel.close();
		} catch (IOException e) {
			log.warn("close", e);
		}
	}

	private CompletableFuture<Void> transferFile(final String resource,final SocketChannel channel) {
		//task: #2
		return CompletableFuture.runAsync(() -> doTransferFile(resource, channel));
	}

	//task: #1
	private void doTransferFile(final String resource, SocketChannel channel) {
		Path filePath = getFilePath(resource);
		try(RandomAccessFile raf = new RandomAccessFile(filePath.toFile(), "rw"); FileChannel fc = raf.getChannel()) {
			writeStatus(Status.OK, channel);
			fc.transferTo(0, raf.length(),channel);
		} catch (FileNotFoundException e) {
			writeStatus(Status.NOT_FOUND, channel);
			writeString("'%s' not found".formatted(resource), channel);
		} catch (IOException e) {
			log.error("read file:", e);
		}
	}

	private Path getFilePath(String resource) {
		//remove trailing '/' to relative path
		String resourcePath;
		if (resource.length() > 1) {
			resourcePath = resource.substring(1);
		} else {
			resourcePath = resource;
		}
		return root.resolve(Paths.get(resourcePath));
	}

	private String getRequest(SocketChannel channel) throws IOException {
		channel.read(buffer);
		buffer.flip();
		String result = new String(buffer.array(), 0, buffer.limit());	
		buffer.clear();
		return result;
	}

	private void writeStatus(Status status,SocketChannel channel) {
		writeString("HTTP/1.1 %d %s".formatted(status.getStatusCode(), status.getReasonPhrase()), channel);
		writeString("", channel);
	}

	private void writeString(String value, SocketChannel socketChannel) {
		buffer.put(value.getBytes());
		buffer.put("\n".getBytes());
		buffer.flip();
		try {
			socketChannel.write(buffer);
		} catch (IOException e) {
			log.error("write: " + value, e);
		}
		buffer.clear();
	}

	private static String getResource(String readedData) {
		try (Scanner scanner = new Scanner(readedData)) {
			scanner.useDelimiter("\\s");
			String method = scanner.next();
			if (!method.equals("GET")) {
				throw new UnsupportedOperationException(method);
			}
			return scanner.next();
		}
	}
}
